use bloomfilter::Bloom;
use clap::{arg, Parser};
use serde_derive::Deserialize;
use std::error::Error;
use std::fs::File;
use std::io::{BufWriter, Write};
use std::path::PathBuf;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Opts {
    /// See CSV record type for format details
    #[arg(short, long, help = "Input file (mnemonics csv)")]
    input_file: PathBuf,

    #[arg(short, long, help = "Output file (bloom filter dump of sha256 hashes)")]
    output_file: PathBuf,

    /// expected during initial use case: 3x 2^32 entries
    /// consisting of the 128, 192 and 256 bit bx PRNG ranges with 32 bit complexity each
    #[arg(
        long,
        help = "Bloom filter: number of items",
        default_value_t = 12_884_901_888
    )]
    num_items: usize,

    /// highly depends on use case
    #[arg(
        long,
        help = "Bloom filter: acceptable rate of false positives",
        default_value_t = 0.000_001
    )]
    fp_rate: f64,
}

#[derive(Debug, Deserialize)]
struct MnemonicsRecord {
    _index: u32,
    mnemonic: String,
}

fn main() -> Result<(), Box<dyn Error>> {
    let opts: Opts = Opts::parse();

    let mut bloom = Bloom::new_for_fp_rate(opts.num_items, opts.fp_rate);

    // Improvement idea: use higher than default capacity?
    let rdr = csv::ReaderBuilder::new()
        .has_headers(false)
        .from_path(&opts.input_file)
        .expect("Failed to open input_file");

    let wtr = File::create(&opts.output_file).expect("error opening output file");
    let mut wtr = BufWriter::new(wtr);

    let mut count: usize = 0;

    rdr.into_deserialize()
        .for_each(|result: Result<MnemonicsRecord, csv::Error>| {
            if count & 0b1_1111_1111_1111_1111_1111 == 0 {
                println!("Read {} lines", count);
            }
            count += 1;

            match result {
                Ok(record) => {
                    let val: String = sha256::digest(record.mnemonic);

                    bloom.set(&val)
                }
                Err(err) => eprintln!("Error reading CSV record: {}", err),
            }
        });

    // Dump bloom filter to file
    println!("Serializing bloom filter to output file");

    // - metadata
    wtr.write_all(&bloom.number_of_bits().to_be_bytes())?;
    wtr.write_all(&bloom.number_of_hash_functions().to_be_bytes())?;
    wtr.write_all(&bloom.sip_keys()[0].0.to_be_bytes())?;
    wtr.write_all(&bloom.sip_keys()[0].1.to_be_bytes())?;
    wtr.write_all(&bloom.sip_keys()[1].0.to_be_bytes())?;
    wtr.write_all(&bloom.sip_keys()[1].1.to_be_bytes())?;

    wtr.flush()?;

    // - bitmap
    wtr.write_all(&bloom.bitmap())?;

    wtr.flush()?;

    Ok(())
}
